import datetime as dt
import re
import tempfile
from os import PathLike, path, listdir
from typing import Union

import pandas as pd
import patoolib
import requests
from bs4 import BeautifulSoup
from pandas.tseries.offsets import MonthEnd

from econuy.resources import updates, columns
from econuy.resources.lstrings import fiscal_url, fiscal_sheets


def get(update_dir: Union[str, PathLike, bool] = False, revise_rows: int = 0,
        save_dir: Union[str, PathLike, bool] = False,
        force_update: bool = False):
    """Get fiscal data.

    Parameters
    ----------
    update_dir : str, PathLike or bool (default is False)
        Path, path-like string pointing to a directory for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save_dir : str, PathLike or bool (default is False)
        Path, path-like string pointing to a directory for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for fiscal data, 25 days)

    Returns
    -------
    output : dictionary of Pandas dataframes
        Each dataframe corresponds to a sheet in the fiscal accounts Excel
        provided by the MEF. Not all sheets are considered.

    """
    update_threshold = 25

    if update_dir is not False:
        update_path = updates._paths(update_dir, multiple=True,
                                     multname="fiscal_nfps")
        modified = dt.datetime.fromtimestamp(path.getmtime(update_path))
        delta = (dt.datetime.now() - modified).days

        if delta < update_threshold and force_update is False:
            print(f"Fiscal data was modified within {update_threshold} day(s)."
                  f" Skipping download...")
            output = {}
            for metadata in fiscal_sheets.values():
                update_path = updates._paths(update_dir, multiple=True,
                                             multname=metadata["Name"])
                delta, previous_data = updates._check_modified(update_path)
                output.update({metadata["Name"]: previous_data})
            return output

    response = requests.get(fiscal_url)
    soup = BeautifulSoup(response.content, "html.parser")
    links = soup.find_all(href=re.compile("\\.rar$"))
    rar = links[0]["href"]
    temp_rar = tempfile.NamedTemporaryFile(suffix=".rar").name
    with open(temp_rar, "wb") as f:
        f.write(requests.get(rar).content)

    with tempfile.TemporaryDirectory() as temp_dir:
        patoolib.extract_archive(temp_rar, outdir=temp_dir, verbosity=-1)
        path_temp = path.join(temp_dir, listdir(temp_dir)[0])

        output = {}
        with pd.ExcelFile(path_temp) as xls:
            for sheet, metadata in fiscal_sheets.items():
                data = (pd.read_excel(xls, sheet_name=sheet).
                        dropna(axis=0, thresh=4).dropna(axis=1, thresh=4).
                        transpose().set_index(2, drop=True))
                data.columns = data.iloc[0]
                data = data[data.index.notnull()].rename_axis(None)
                data.index = data.index + MonthEnd(1)
                data.columns = metadata["Colnames"]

                if update_dir is not False:
                    update_path = updates._paths(update_dir, multiple=True,
                                                 multname=metadata["Name"])
                    delta, previous_data = updates._check_modified(update_path)
                    data = updates._revise(new_data=data,
                                           prev_data=previous_data,
                                           revise_rows=revise_rows)
                data = data.apply(pd.to_numeric, errors="coerce")
                columns._setmeta(
                    data, area="Cuentas fiscales y deuda", currency="UYU",
                    inf_adj="No", index="No", seas_adj="NSA", ts_type="Flujo",
                    cumperiods=1
                )

                if save_dir is not False:
                    save_path = updates._paths(save_dir, multiple=True,
                                               multname=metadata["Name"])
                    data.to_csv(save_path)

                output.update({metadata["Name"]: data})

    return output



import datetime as dt
from os import PathLike
from typing import Union

import numpy as np
import pandas as pd
import requests
from pandas.tseries.offsets import MonthEnd

from econuy.processing import index
from econuy.resources import updates, columns
from econuy.resources.lstrings import reer_url, ar_cpi_url, ar_cpi_payload
from econuy.retrieval import cpi, nxr


def get_official(update: Union[str, PathLike, bool] = False,
                 revise_rows: int = 0, save: Union[str, PathLike, bool] = False,
                 force_update: bool = False):
    """Get official real exchange rates from the BCU website.

    Parameters
    ----------
    update : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for RXR, 25 days)

    Returns
    -------
    proc : Pandas dataframe

    """
    update_threshold = 25

    if update is not None:
        update_path = updates._paths(update, multiple=False,
                                     name="rxr_official.csv")
        delta, previous_data = updates._check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    raw = pd.read_excel(reer_url, skiprows=8, usecols="B:H", index_col=0)
    proc = raw.dropna(how="any")
    proc.columns = ["Global", "Regional", "Extrarregional",
                    "Argentina", "Brasil", "EEUU"]
    proc.index = pd.to_datetime(proc.index) + MonthEnd(1)

    if update is not False:
        proc = updates._revise(new_data=proc, prev_data=previous_data,
                               revise_rows=revise_rows)

    columns._setmeta(proc, area="Precios y salarios", currency="-",
                     inf_adj="No", index="2017", seas_adj="NSA",
                     ts_type="-", cumperiods=1)

    if save is not None:
        save_path = updates._paths(save, multiple=False,
                                   name="rxr_official.csv")
        proc.to_csv(save_path)

    return proc


def get_custom(update: Union[str, PathLike, bool] = False,
               revise_rows: int = 0, save: Union[str, PathLike, bool] = False,
               force_update: bool = False):
    """Calculate custom real exchange rates from various sources.

    Parameters
    ----------
    update : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for RXR, 25 days)

    Returns
    -------
    output : Pandas dataframe

    """
    update_threshold = 25

    if update is not None:
        update_path = updates._paths(update, multiple=False,
                                     name="rxr_custom.csv")
        delta, previous_data = updates._check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    url_ = "http://dataservices.imf.org/REST/SDMX_JSON.svc/CompactData/IFS/M."
    url_extra = ".?startPeriod=1970&endPeriod="
    raw = []
    for country in ["US", "BR", "AR"]:
        for indicator in ["PCPI_IX", "ENDA_XDC_USD_RATE"]:
            base_url = (f"{url_}{country}.{indicator}{url_extra}"
                        f"{dt.datetime.now().year}")
            r_json = requests.get(base_url).json()
            data = r_json["CompactData"]["DataSet"]["Series"]["Obs"]
            try:
                data = pd.DataFrame(data)
                data.set_index("@TIME_PERIOD", drop=True, inplace=True)
            except ValueError:
                data = pd.DataFrame(np.nan,
                                    index=pd.date_range(start="1970-01-01",
                                                        end=dt.datetime.now(),
                                                        freq="M"),
                                    columns=[f"{country}.{indicator}"])
            if "@OBS_STATUS" in data.columns:
                data.drop("@OBS_STATUS", inplace=True, axis=1)
            data.index = (pd.to_datetime(data.index, format="%Y-%m")
                          + MonthEnd(1))
            data.columns = [f"{country}.{indicator}"]
            raw.append(data)
    raw = pd.concat(raw, axis=1, sort=True).apply(pd.to_numeric)

    ar_black_xr, ar_cpi = missing_ar()
    proc = raw.copy()
    proc["AR.PCPI_IX"] = ar_cpi
    ar_black_xr = pd.concat([ar_black_xr, proc["AR.ENDA_XDC_USD_RATE"]], axis=1)
    ar_black_xr[0] = np.where(pd.isna(ar_black_xr[0]),
                              ar_black_xr["AR.ENDA_XDC_USD_RATE"],
                              ar_black_xr[0])
    proc["AR.ENDA_XDC_USD_RATE_black"] = ar_black_xr.iloc[:, 0]
    proc["AR_E_A"] = proc.iloc[:, [5, 6]].mean(axis=1)

    uy_cpi = cpi.get(update="cpi.csv", revise_rows=6,
                     save="cpi.csv", force_update=False)
    uy_e = nxr.get(update="nxr.csv", revise_rows=6,
                   save="nxr.csv", force_update=False).iloc[:, [3]]
    proc = pd.concat([proc, uy_cpi, uy_e], axis=1)
    proc = proc.interpolate(method="linear", limit_area="inside")
    proc = proc.dropna(how="any")
    proc.columns = ["US_P", "US_E", "BR_P", "BR_E", "AR_P", "AR_E",
                    "AR_E_B", "AR_E_A", "UY_P", "UY_E"]

    output = pd.DataFrame()
    output["UY_E_P"] = proc["UY_E"] / proc["UY_P"]
    output["TCR_UY_AR"] = output["UY_E_P"] / proc["AR_E_A"] * proc["AR_P"]
    output["TCR_UY_BR"] = output["UY_E_P"] / proc["BR_E"] * proc["BR_P"]
    output["TCR_UY_US"] = output["UY_E_P"] * proc["US_P"]
    output["TCR_AR_US"] = proc["BR_E"] * proc["US_P"] / proc["BR_P"]
    output["TCR_BR_US"] = proc["AR_E"] * proc["US_P"] / proc["AR_P"]
    output.drop("UY_E_P", axis=1, inplace=True)

    columns._setmeta(output, area="Precios y salarios", currency="-",
                     inf_adj="-", index="-", seas_adj="NSA",
                     ts_type="Flujo", cumperiods=1)
    output = index.base_index(output, start_date="2010-01-01",
                              end_date="2010-12-31", base=100)

    if update is not False:
        output = updates._revise(new_data=proc, prev_data=previous_data,
                                 revise_rows=revise_rows)

    if save is not False:
        save_path = updates._paths(save, multiple=False,
                                   name="rxr_custom.csv")
        output.to_csv(save_path, sep=" ")

    return output


def missing_ar():
    """Get Argentina's non-official exchange rate and CPI."""
    today_fmt = dt.datetime.now().strftime("%d-%m-%Y")
    black_url = (f"https://mercados.ambito.com/dolar/informal/"
                 f"historico-general/11-01-2002/{today_fmt}")
    black_r = requests.get(black_url).json()
    black_xr = pd.DataFrame(black_r)
    black_xr.set_index(0, drop=True, inplace=True)
    black_xr.drop("Fecha", inplace=True)
    black_xr = black_xr.replace(",", ".", regex=True).apply(pd.to_numeric)
    black_xr.index = pd.to_datetime(black_xr.index, format="%d-%m-%Y")
    black_xr = black_xr.mean(axis=1).to_frame().sort_index()
    black_xr = black_xr.resample("M").mean()

    cpi_r = requests.get(ar_cpi_url, params=ar_cpi_payload)
    cpi_ar = pd.read_html(cpi_r.content)[0]
    cpi_ar.set_index("Fecha", drop=True, inplace=True)
    cpi_ar.index = pd.to_datetime(cpi_ar.index, format="%d/%m/%Y")
    cpi_ar.columns = ["nivel"]
    cpi_ar = cpi_ar.divide(10)

    ps_url = "http://www.inflacionverdadera.com/Argentina_inflation.xls"
    cpi_ps = pd.read_excel(ps_url)
    cpi_ps.set_index("date", drop=True, inplace=True)
    cpi_ps.index = cpi_ps.index + MonthEnd(1)
    cpi_ps = cpi_ps.loc[(cpi_ps.index >= "2006-12-31") &
                        (cpi_ps.index <= "2016-12-01"), "index"]
    cpi_ps = cpi_ps.to_frame().pct_change(periods=1).multiply(100)
    cpi_ps.columns = ["nivel"]

    cpi_all = (cpi_ar.append(cpi_ps).reset_index().
               drop_duplicates(subset="index", keep="last").
               set_index("index", drop=True).sort_index())

    cpi_all = cpi_all.divide(100).add(1).cumprod()

    return black_xr, cpi_all



import datetime as dt
from os import PathLike
from typing import Union

import pandas as pd
from pandas.tseries.offsets import MonthEnd

from econuy.processing import freqs, convert
from econuy.resources import updates, columns
from econuy.resources.lstrings import nat_accounts_metadata


def get(update_dir: Union[str, PathLike, bool] = False, revise_rows: int = 0,
        save_dir: Union[str, PathLike, bool] = False,
        force_update: bool = False):
    """Get national accounts data.

    Parameters
    ----------
    update_dir : str, PathLike or bool (default is False)
        Path, path-like string pointing to a directory for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save_dir : str, PathLike or bool (default is False)
        Path, path-like string pointing to a directory for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for national accounts, 80 days)

    Returns
    -------
    parsed_excels : dictionary of Pandas dataframes
        Each dataframe corresponds to a national accounts table.

    """
    update_threshold = 80

    parsed_excels = {}
    for file, metadata in nat_accounts_metadata.items():

        if update_dir is not False:
            update_path = updates._paths(update_dir, multiple=True,
                                         multname=metadata["Name"])
            delta, previous_data = updates._check_modified(update_path)

            if delta < update_threshold and force_update is False:
                print(f"{metadata['Name']}.csv was modified within"
                      f" {update_threshold} day(s). Skipping download...")
                parsed_excels.update({metadata["Name"]: previous_data})
                continue

        raw = pd.read_excel(file, skiprows=9, nrows=metadata["Rows"])
        proc = (raw.drop(columns=["Unnamed: 0"]).
                dropna(axis=0, how="all").dropna(axis=1, how="all"))
        proc = proc.transpose()
        proc.columns = metadata["Colnames"]
        proc.drop(["Unnamed: 1"], inplace=True)

        _fix_dates(proc)

        if metadata["Index"] == "No":
            proc = proc.divide(1000)
        if update_dir is not False:
            proc = updates._revise(new_data=proc, prev_data=previous_data,
                                   revise_rows=revise_rows)
        proc = proc.apply(pd.to_numeric, errors="coerce")

        columns._setmeta(proc, area="Actividad económica", currency="UYU",
                         inf_adj=metadata["Inf. Adj."],
                         index=metadata["Index"],
                         seas_adj=metadata["Seas"], ts_type="Flujo",
                         cumperiods=1)

        if save_dir is not False:
            save_path = updates._paths(save_dir, multiple=True,
                                       multname=metadata["Name"])
            proc.to_csv(save_path)

        parsed_excels.update({metadata["Name"]: proc})

    return parsed_excels


def _fix_dates(df):
    """Cleanup dates inplace in BCU national accounts files."""
    df.index = df.index.str.replace("*", "")
    df.index = df.index.str.replace(r"\bI \b", "3-", regex=True)
    df.index = df.index.str.replace(r"\bII \b", "6-", regex=True)
    df.index = df.index.str.replace(r"\bIII \b", "9-", regex=True)
    df.index = df.index.str.replace(r"\bIV \b", "12-", regex=True)
    df.index = pd.to_datetime(df.index, format="%m-%Y") + MonthEnd(1)


def _lin_gdp(update: Union[str, PathLike, bool] = False,
             save: Union[str, PathLike, bool] = False,
             force_update: bool = False):
    """Get nominal GDP data in UYU and USD with forecasts.

    Update nominal GDP data for use in the `convert.pcgdp()` function.
    Get IMF forecasts for year of last available data point and the next
    year (for example, if the last period available at the BCU website is
    september 2019, fetch forecasts for 2019 and 2020).

    Parameters
    ----------
    update : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't saving.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for national accounts data, 80 days)

    Returns
    -------
    output : Pandas dataframe
        Quarterly GDP in UYU and USD with 1 year forecasts.

    """
    update_threshold = 80

    if update is not False:
        update_path = updates._paths(update, multiple=False, name="lin_gdp.csv")
        delta, previous_data = updates._check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    data_uyu = get(update_dir=True, revise_rows=4, save_dir=True,
                   force_update=False)["na_gdp_cur_nsa"]
    data_uyu = freqs.rolling(data_uyu, periods=4, operation="sum")
    data_usd = convert.usd(data_uyu)

    data = [data_uyu, data_usd]
    last_year = data_uyu.index.max().year
    if data_uyu.index.max().month == 12:
        last_year += 1

    results = []
    for table, gdp in zip(["NGDP", "NGDPD"], data):

        table_url = (f"https://www.imf.org/external/pubs/ft/weo/2019/02/weodat"
                     f"a/weorept.aspx?sy={last_year-1}&ey={last_year+1}&scsm=1"
                     f"&ssd=1&sort=country&ds=.&br=1&pr1.x=27&pr1.y=9&c=298&s"
                     f"={table}&grp=0&a=")
        imf_data = pd.to_numeric(pd.read_html(table_url)[4].iloc[2, [5, 6, 7]])
        imf_data = imf_data.reset_index(drop=True)
        fcast = (gdp.loc[[dt.datetime(last_year-1, 12, 31)]].
                 multiply(imf_data.iloc[1]).divide(imf_data.iloc[0]))
        fcast = fcast.rename(index={dt.datetime(last_year-1, 12, 31):
                                    dt.datetime(last_year, 12, 31)})
        next_fcast = (gdp.loc[[dt.datetime(last_year-1, 12, 31)]].
                      multiply(imf_data.iloc[2]).divide(imf_data.iloc[0]))
        next_fcast = next_fcast.rename(index={dt.datetime(last_year-1, 12, 31):
                                              dt.datetime(last_year+1, 12, 31)})
        fcast = fcast.append(next_fcast)
        gdp = gdp.append(fcast)
        results.append(gdp)

    output = pd.concat(results, axis=1)
    output = output.resample("Q-DEC").interpolate("linear")

    if save is not False:
        save_path = updates._paths(save, multiple=False, name="lin_gdp.csv")
        output.to_csv(save_path)

    return output



from os import PathLike
from typing import Union

import pandas as pd
from pandas.tseries.offsets import MonthEnd

from econuy.resources import updates, columns
from econuy.resources.lstrings import cpi_url


def get(update: Union[str, PathLike, bool] = False, revise_rows: int = 0,
        save: Union[str, PathLike, bool] = False, force_update: bool = False):
    """Get CPI data.

    Parameters
    ----------
    update : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for CPI, 25 days)

    Returns
    -------
    cpi : Pandas dataframe

    """
    update_threshold = 25

    if update is not False:
        update_path = updates._paths(update, multiple=False, name="cpi.csv")
        delta, previous_data = updates._check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    cpi_raw = pd.read_excel(cpi_url, skiprows=7).dropna(axis=0, thresh=2)
    cpi = (cpi_raw.drop(["Mensual", "Acum.año", "Acum.12 meses"], axis=1).
           dropna(axis=0, how="all").set_index("Mes y año").rename_axis(None))
    cpi.columns = ["Índice de precios al consumo"]
    cpi.index = cpi.index + MonthEnd(1)

    if update is not False:
        cpi = updates._revise(new_data=cpi, prev_data=previous_data,
                              revise_rows=revise_rows)

    cpi = cpi.apply(pd.to_numeric, errors="coerce")
    columns._setmeta(cpi, area="Precios y salarios", currency="-",
                     inf_adj="No", index="2010-10-31", seas_adj="NSA",
                     ts_type="-", cumperiods=1)

    if save is not False:
        save_path = updates._paths(save, multiple=False, name="cpi.csv")
        cpi.to_csv(save_path)

    return cpi



from os import PathLike
from typing import Union

import pandas as pd
from pandas.tseries.offsets import MonthEnd

from econuy.resources import updates, columns
from econuy.resources.lstrings import nxr_url


def get(update: Union[str, PathLike, bool] = False, revise_rows: int = 0,
        save: Union[str, PathLike, bool] = False, force_update: bool = False):
    """Get nominal exchange rate data.

    Parameters
    ----------
    update : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for NXR, 25 days)

    Returns
    -------
    nxr : Pandas dataframe

    """
    update_threshold = 25

    if update is not False:
        update_path = updates._paths(update, multiple=False, name="nxr.csv")
        delta, previous_data = updates._check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    nxr_raw = pd.read_excel(nxr_url, skiprows=4)
    nxr = (nxr_raw.dropna(axis=0, thresh=4).set_index("Mes y año").
           dropna(axis=1, how="all").rename_axis(None))
    nxr.columns = ["Tipo de cambio compra, fin de período",
                   "Tipo de cambio venta, fin de período",
                   "Tipo de cambio compra, promedio",
                   "Tipo de cambio venta, promedio"]
    nxr.index = nxr.index + MonthEnd(1)

    if update is not False:
        nxr = updates._revise(new_data=nxr, prev_data=previous_data,
                              revise_rows=revise_rows)

    nxr = nxr.apply(pd.to_numeric, errors="coerce")
    columns._setmeta(nxr, area="Precios y salarios", currency="-",
                     inf_adj="No", index="No", seas_adj="NSA",
                     ts_type="-", cumperiods=1)

    if save is not False:
        save_path = updates._paths(save, multiple=False, name="nxr.csv")
        nxr.to_csv(save_path)

    return nxr



from os import PathLike
from typing import Union

import pandas as pd

from econuy.resources import columns, updates
from econuy.retrieval import fx_ff, reserves_chg


def get(save: Union[str, PathLike, bool] = False):
    """Get spot, future and forwards FX operations by the Central Bank."""
    changes = reserves_chg.get(update=True, save=True)
    ff = fx_ff.get(update=True, save=True)
    spot = changes.iloc[:, 0]
    fx_ops = pd.merge(spot, ff, how="outer", left_index=True, right_index=True)
    fx_ops = fx_ops.loc[(fx_ops.index >= ff.index.min()) &
                        (fx_ops.index <= spot.index.max())]
    fx_ops = fx_ops.apply(pd.to_numeric, errors="coerce")
    fx_ops = fx_ops.fillna(0)
    fx_ops.columns = ["Spot", "Futuros", "Forwards"]

    columns._setmeta(fx_ops, area="Reservas internacionales",
                     currency="USD", inf_adj="No", index="No",
                     seas_adj="NSA", ts_type="Flujo", cumperiods=1)

    if save is not False:
        save_path = updates._paths(save, multiple=False,
                                   name="fx_spot_ff.csv")
        fx_ops.to_csv(save_path)

    return fx_ops



import datetime as dt
import urllib
from os import PathLike
from typing import List, Union

import pandas as pd
from dateutil.relativedelta import relativedelta

from econuy.resources import columns, updates
from econuy.resources.lstrings import (reserves_cols, reserves_url,
                                       missing_reserves_url)

months = ["ene", "feb", "mar", "abr", "may", "jun",
          "jul", "ago", "set", "oct", "nov", "dic"]
years = list(range(2013, dt.datetime.now().year + 1))
files_ = [month + str(year) for year in years for month in months]


def get(files: List[str] = files_, update: Union[str, PathLike, bool] = False,
        save: Union[str, PathLike, bool] = False):
    """Get international reserves change data from online sources.

    Use as input a list of strings of the format %b%Y, each representing a
    month of data.

    Parameters
    ----------
    files : list of strings
        List of strings of the type '%b%Y', only %b is in Spanish. So 'ene'
        instead of 'jan'. For example, 'oct2017'.
    update : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.

    Returns
    -------
    reserves : Pandas dataframe

    """
    urls = [f"{reserves_url}{file}.xls" for file in files]
    wrong_may14 = f"{reserves_url}may2014.xls"
    fixed_may14 = f"{reserves_url}mayo2014.xls"
    urls = [fixed_may14 if x == wrong_may14 else x for x in urls]

    if update is not False:
        update_path = updates._paths(update, multiple=False,
                                     name="reserves_chg.csv")
        previous_data = pd.read_csv(update_path, index_col=0,
                                    header=list(range(9)))
        previous_data.columns = reserves_cols[1:46]
        previous_data.index = pd.to_datetime(previous_data.index)
        urls = urls[-18:]

    reports = []
    for url in urls:

        try:
            with pd.ExcelFile(url) as xls:
                month_of_report = pd.read_excel(xls, sheet_name="INDICE")
                raw = pd.read_excel(xls, sheet_name="ACTIVOS DE RESERVA",
                                    skiprows=3)
            first_day = month_of_report.iloc[7, 4]
            last_day = (first_day
                        + relativedelta(months=1)
                        - dt.timedelta(days=1))
            proc = raw.dropna(axis=0, thresh=20).dropna(axis=1, thresh=20)
            proc = proc.transpose()
            proc.index.name = "Date"
            proc = proc.iloc[:, 1:46]
            proc.columns = reserves_cols[1:46]
            proc = proc.iloc[1:]
            proc.index = pd.to_datetime(proc.index, errors="coerce")
            proc = proc.loc[proc.index.dropna()]
            proc = proc.loc[first_day:last_day]
            reports.append(proc)

        except urllib.error.HTTPError:
            print(f"{url} could not be reached.")
            pass

    mar14 = pd.read_excel(missing_reserves_url, index_col=0)
    mar14.columns = reserves_cols[1:46]
    reserves = pd.concat(reports + [mar14], sort=False).sort_index()

    if update is not False:
        reserves = previous_data.append(reserves, sort=False)
        reserves = reserves.loc[~reserves.index.duplicated(keep="last")]

    reserves = reserves.apply(pd.to_numeric, errors="coerce")
    columns._setmeta(reserves, area="Reservas internacionales",
                     currency="USD", inf_adj="No", index="No",
                     seas_adj="NSA", ts_type="Flujo", cumperiods=1)

    if save is not False:
        save_path = updates._paths(save, multiple=False,
                                   name="reserves_chg.csv")
        reserves.to_csv(save_path)

    return reserves



from os import PathLike
from typing import Union

import pandas as pd

from econuy.resources import updates, columns
from econuy.resources.lstrings import labor_url


def get(update: Union[str, PathLike, bool] = False, revise_rows: int = 0,
        save: Union[str, PathLike, bool] = False, force_update: bool = False):
    """Get labor market data.

    Get monthly labor force participation rate, employment rate (employment to
    working-age population) and unemployment rate.

    Parameters
    ----------
    update : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for labor market, 25 days)

    Returns
    -------
    labor : Pandas dataframe

    """
    update_threshold = 25

    if update is not False:
        update_path = updates._paths(update, multiple=False, name="labor.csv")
        delta, previous_data = updates._check_modified(update_path)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    labor_raw = pd.read_excel(labor_url, skiprows=39).dropna(axis=0, thresh=2)
    labor = labor_raw[~labor_raw["Unnamed: 0"].str.contains("-|/|Total",
                                                            regex=True)]
    labor = labor[["Unnamed: 1", "Unnamed: 4", "Unnamed: 7"]]
    labor.index = pd.date_range(start="2006-01-01",
                                periods=len(labor), freq="M")
    labor.columns = ["Tasa de actividad", "Tasa de empleo",
                     "Tasa de desempleo"]

    if update is not False:
        labor = updates._revise(new_data=labor, prev_data=previous_data,
                                revise_rows=revise_rows)

    labor = labor.apply(pd.to_numeric, errors="coerce")
    columns._setmeta(labor, area="Mercado laboral", currency="-",
                     inf_adj="No", index="No", seas_adj="NSA",
                     ts_type="-", cumperiods=1)

    if save is not False:
        save_path = updates._paths(save, multiple=False, name="labor.csv")
        labor.to_csv(save_path)

    return labor



import datetime as dt
import tempfile
import zipfile
from io import BytesIO
from os import PathLike, path
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import requests
from pandas.tseries.offsets import YearEnd

from econuy.resources import updates
from econuy.resources.lstrings import (beef_url, pulp_url, soybean_url,
                                       what_url, imf_url, milk1_url, milk2_url)


def _weights(update: Union[str, PathLike, bool] = False, revise_rows: int = 0,
             save: Union[str, PathLike, bool] = False,
             force_update: bool = False):
    """Get weights for the commodity price index from the UN COMTRADE database.

    Parameters
    ----------
    update : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for commodity weights, 85 days)

    Returns
    -------
    output : Pandas dataframe
        Export-based weights for relevant commodities to Uruguay.

    """
    update_threshold = 85

    if update is not False:
        update_path = updates._paths(update, multiple=False,
                                     name="comm_weights.csv")
        delta, previous_data = updates._check_modified(update_path,
                                                       multiindex=False)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    raw = []
    prods = "%2C".join(["0011", "011", "01251", "01252", "0176", "022", "041",
                        "042", "043", "2222", "24", "25", "268", "97"])
    base_url = "http://comtrade.un.org/api/get?max=1000&type=C&freq=A&px=S3&ps"
    for year in range(1992, dt.datetime.now().year-1):
        full_url = f"{base_url}={year}&r=all&p=858&rg=1&cc={prods}"
        un_r = requests.get(full_url)
        raw.append(pd.DataFrame(un_r.json()["dataset"]))
    raw = pd.concat(raw, axis=0)

    table = raw.groupby(["period", "cmdDescE"]).sum().reset_index()
    table = table.pivot(index="period", columns="cmdDescE", values="TradeValue")
    table.fillna(0, inplace=True)
    percentage = table.div(table.sum(axis=1), axis=0)
    percentage.index = (pd.to_datetime(percentage.index, format="%Y")
                        + YearEnd(1))
    rolling = percentage.rolling(window=3, min_periods=3).mean()
    output = rolling.resample("M").bfill()

    beef = ["BOVINE MEAT", "Edible offal of bovine animals, fresh or chilled",
            "Meat and offal (other than liver), of bovine animals, "
            "prepared or preserv", "Edible offal of bovine animals, frozen",
            "Bovine animals, live"]
    output["Beef"] = output[beef].sum(axis=1, min_count=len(beef))
    output.drop(beef, axis=1, inplace=True)
    output.columns = ["Barley", "Wood", "Gold", "Milk", "Pulp",
                      "Rice", "Soybeans", "Wheat", "Wool", "Beef"]

    if update is not False:
        output = updates._revise(new_data=output, prev_data=previous_data,
                                 revise_rows=revise_rows)

    if save is not False:
        save_path = updates._paths(save, multiple=False,
                                   name="comm_weights.csv")
        output.to_csv(save_path)

    return output


def _prices(update: Union[str, Path, bool] = False, revise_rows: int = 0,
            save: Union[str, Path, bool] = False, force_update: bool = False):
    """Prepare prices for the commodity price index.

    Parameters
    ----------
    update : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for updating, or bool,
        in which case if True, save in predefined file, or False, don't update.
    revise_rows : int (default is 0)
        How many rows of old data to replace with new data.
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    force_update : bool (default is False)
        If True, fetch data and update existing data even if it was modified
        within its update window (for commodity prices, 10 days)

    Returns
    -------
    complete : Pandas dataframe
        Prices and price indexes of relevant commodities for Uruguay.

    """
    update_threshold = 10
    bushel_conv = 36.74 / 100

    if update is not False:
        update_path = updates._paths(update, multiple=False,
                                     name="comm_prices.csv")
        delta, previous_data = updates._check_modified(update_path,
                                                       multiindex=False)

        if delta < update_threshold and force_update is False:
            print(f"{update} was modified within {update_threshold} day(s). "
                  f"Skipping download...")
            return previous_data

    raw_beef = (pd.read_excel(beef_url, header=4, index_col=0)
                .dropna(how="all").drop(index=0))
    raw_beef.columns = raw_beef.columns.str.strip()
    proc_beef = raw_beef["Ing. Prom./Ton."].to_frame()
    proc_beef.index = pd.date_range(start="2002-01-04",
                                    periods=len(proc_beef), freq="W-SAT")
    proc_beef["Ing. Prom./Ton."] = np.where(
        proc_beef > np.mean(proc_beef) + np.std(proc_beef) * 2,
        proc_beef / 1000,
        proc_beef,
    )
    beef = proc_beef.resample("M").mean()

    raw_pulp_r = requests.get(pulp_url)
    temp_dir = tempfile.TemporaryDirectory()
    with zipfile.ZipFile(BytesIO(raw_pulp_r.content), "r") as f:
        f.extractall(path=temp_dir.name)
        path_temp = path.join(temp_dir.name, "monthly_values.csv")
        raw_pulp = pd.read_csv(path_temp, sep=";").dropna(how="any")
    proc_pulp = raw_pulp.copy().sort_index(ascending=False)
    proc_pulp.index = pd.date_range(start="1990-01-31",
                                    periods=len(proc_pulp), freq="M")
    proc_pulp.drop(["Label", "Codes"], axis=1, inplace=True)
    pulp = proc_pulp

    soy_wheat = []
    for url in [soybean_url, what_url]:
        raw = pd.read_csv(url, index_col=0)
        proc = (raw["Settle"] * bushel_conv).to_frame()
        proc.index = pd.to_datetime(proc.index, format="%Y-%m-%d")
        proc.sort_index(inplace=True)
        soy_wheat.append(proc.resample("M").mean())
    soybean = soy_wheat[0]
    wheat = soy_wheat[1]

    raw_milk = pd.read_excel(milk1_url, skiprows=13,
                             nrows=dt.datetime.now().year - 2007)
    raw_milk.dropna(how="all", axis=1, inplace=True)
    raw_milk.drop(["Promedio ", "Variación"], axis=1, inplace=True)
    raw_milk.columns = ["Año/Mes"] + list(range(1, 13))
    proc_milk = pd.melt(raw_milk, id_vars=["Año/Mes"])
    proc_milk.sort_values(by=["Año/Mes", "variable"], inplace=True)
    proc_milk.index = pd.date_range(start="2007-01-31",
                                    periods=len(proc_milk), freq="M")
    proc_milk = proc_milk.iloc[:, 2].to_frame()

    prev_milk = pd.read_excel(milk2_url, sheet_name="Dairy Products Prices",
                              index_col=0, usecols="A,D", skiprows=5)
    prev_milk = prev_milk.resample("M").mean()
    eurusd_r = requests.get(
        "http://fx.sauder.ubc.ca/cgi/fxdata",
        params=f"b=USD&c=EUR&rd=&fd=1&fm=1&fy=2001&ld=31&lm=12&ly="
               f"{dt.datetime.now().year}&y=monthly&q=volume&f=html&o=&cu=on"
    )
    eurusd = pd.read_html(eurusd_r.content)[0].drop("MMM YYYY", axis=1)
    eurusd.index = pd.date_range(start="2001-01-31", periods=len(eurusd),
                                 freq="M")
    prev_milk = prev_milk.divide(eurusd.values).multiply(10)
    prev_milk = prev_milk.loc[prev_milk.index < min(proc_milk.index)]
    prev_milk.columns, proc_milk.columns = ["Price"], ["Price"]
    milk = prev_milk.append(proc_milk)

    raw_imf = pd.read_excel(imf_url)
    raw_imf.columns = raw_imf.iloc[0, :]
    proc_imf = raw_imf.iloc[3:, 1:]
    proc_imf.index = pd.date_range(start="1980-01-31",
                                   periods=len(proc_imf), freq="M")
    rice = proc_imf[proc_imf.columns[proc_imf.columns.str.contains("Rice")]]
    wood = proc_imf[proc_imf.columns[
        proc_imf.columns.str.contains("Sawnwood")
    ]]
    wood = wood.mean(axis=1).to_frame()
    wool = proc_imf[proc_imf.columns[proc_imf.columns.str.startswith("Wool")]]
    wool = wool.mean(axis=1).to_frame()
    barley = proc_imf[proc_imf.columns[
        proc_imf.columns.str.startswith("Barley")
    ]]
    gold = proc_imf[proc_imf.columns[
        proc_imf.columns.str.startswith("Gold")
    ]]

    complete = pd.concat([beef, pulp, soybean, milk, rice, wood, wool, barley,
                          gold, wheat], axis=1)
    complete = complete.reindex(beef.index).dropna(thresh=9)
    complete.columns = ["Beef", "Pulp", "Soybeans", "Milk", "Rice", "Wood",
                        "Wool", "Barley", "Gold", "Wheat"]

    if update is not False:
        complete = updates._revise(new_data=complete, prev_data=previous_data,
                                   revise_rows=revise_rows)

    if save is not False:
        save_path = updates._paths(save, multiple=False,
                                   name="comm_prices.csv")
        complete.to_csv(save_path)

    return complete


def get(save: Union[str, Path, bool] = False,
        update_prices: Union[str, Path, bool] = False,
        update_weights: Union[str, Path, bool] = False):
    """Get the commodity price index.

    Parameters
    ----------
    save : str, PathLike or bool (default is False)
        Path, path-like string pointing to a CSV file for saving, or bool,
        in which case if True, save in predefined file, or False, don't save.
    update_prices : str, PathLike or bool (default is False)
        Whether to update commodity prices with the `prices()` function.
    update_weights : str, PathLike or bool (default is False)
        Whether to update commodity weights with the `weights()` function.

    Returns
    -------
    product : Pandas dataframe
        Export-weighted average of commodity prices relevant to Uruguay.

    """
    prices = _prices(update=update_prices, revise_rows=3,
                     save=True, force_update=False)
    prices = prices.interpolate(method="linear", limit=1).dropna(how="any")
    prices = prices.pct_change(periods=1)
    weights = _weights(update=update_weights, revise_rows=24,
                       save=True, force_update=False)
    weights = weights[prices.columns]
    weights = weights.reindex(prices.index, method="ffill")

    product = pd.DataFrame(prices.values * weights.values,
                           columns=prices.columns, index=prices.index)
    product = product.sum(axis=1).add(1).to_frame().cumprod()

    if save is not False:
        save_path = updates._paths(save, multiple=False,
                                   name="commodity_index.csv")
        product.to_csv(save_path)

    return product



import datetime as dt
from os import path, makedirs, PathLike
from pathlib import Path
from typing import Union

import pandas as pd


def _check_modified(data_path: Union[str, PathLike], multiindex=True):
    """Check how long it has been since the file in path was modified."""
    try:
        modified_time = dt.datetime.fromtimestamp(path.getmtime(data_path))
        delta = (dt.datetime.now() - modified_time).days
        if multiindex is True:
            previous_data = pd.read_csv(data_path, index_col=0,
                                        header=list(range(9)))
        else:
            previous_data = pd.read_csv(data_path, index_col=0)
        previous_data.index = pd.to_datetime(previous_data.index)
    except FileNotFoundError:
        print(f"{data_path} does not exist. No data will be updated")
        delta = 0
        previous_data = pd.DataFrame()

    return delta, previous_data


def _revise(new_data: pd.DataFrame, prev_data: pd.DataFrame, revise_rows: int):
    """Replace n rows of data at the end of a dataframe with new data."""
    if len(prev_data) == 0:
        return new_data
    non_revised = prev_data[:len(prev_data)-revise_rows]
    revised = new_data[len(prev_data)-revise_rows:]
    non_revised.columns = new_data.columns
    updated = non_revised.append(revised, sort=False)

    return updated


def _paths(filepath: Union[str, PathLike, bool], multiple: bool = False,
           name: str = None, multname: str = None):
    """Take a path-like object or bool and return a full path."""
    if multiple is False:
        if isinstance(filepath, PathLike) or isinstance(filepath, str):
            final_path = filepath
        else:
            final_path = path.join("econuy-data", name)
    else:
        if isinstance(filepath, PathLike):
            base = Path(filepath).as_posix()
            if path.isfile(base):
                base = path.dirname(base)
            final_path = path.join(base, multname + ".csv")
        elif isinstance(filepath, str):
            base = filepath
            if path.isfile(base):
                base = path.dirname(base)
            final_path = path.join(base,  multname + ".csv")
        else:
            final_path = path.join("econuy-data", multname + ".csv")
    if not path.exists(path.dirname(final_path)):
        makedirs(path.dirname(final_path))

    return final_path


def rsearch(dir_file: Union[str, PathLike], search_term: str, n: int = 2):
    """Recursively search for a file starting from the n-parent folder of
    a supplied path."""
    i = 0
    while i < n:
        i += 1
        dir_file = path.dirname(dir_file)
    try:
        final_path = ([x for x in Path(dir_file).rglob(search_term)][0]
                      .absolute().as_posix())
    except IndexError:
        final_path = True
    return final_path



from os import PathLike, path, getcwd
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
from statsmodels.api import tsa
from statsmodels.tools.sm_exceptions import X13Error

from econuy.resources import columns, updates


def decompose(df: pd.DataFrame, trading: bool = True, outlier: bool = True,
              x13_binary: Union[str, PathLike] = "search",
              search_parents: int = 1):
    """Apply X13 decomposition. Return trend and seasonally adjusted dataframe.

    Decompose the series in a Pandas dataframe using the US Census X13
    methodology. Will try different combinations of the `trading` and `outlier`
    arguments if an X13 error is raised. Requires having the X13 binary in the
    `resources` folder. Please refer to the README for instructions on where
    to get this binary.

    Parameters
    ----------
    df : Pandas dataframe
    trading : bool (default is True)
        Whether to automatically detect trading days.
    outlier : bool (default is True)
        Whether to automatically detect outliers.
    x13_binary: str or PathLike (default is "search")
        Location of the X13 binary. If "search" is used, will attempt to find
        the binary in the project structure (up to two parent folders).
    search_parents: int (default is 2)
        If "search" is chosen for `x13_binary`, this parameter controls how
        many parent directories to go up before recursively searching for the
        binary.

    Returns
    -------
    trend, seas_adj : Pandas dataframe
        Dataframes of the same shape of the input dataframe, containing the
        trend component and the seasonally adjusted series.

    """
    if x13_binary == "search":
        binary_path = updates.rsearch(dir_file=getcwd(), n=search_parents,
                                      search_term="x13*")
    elif isinstance(x13_binary, str):
        binary_path = x13_binary
    else:
        binary_path = Path(x13_binary).as_posix()

    if path.isfile(binary_path) is False:
        print("X13 binary missing. Please refer to the README"
              "for instructions on where to get binaries for Windows and Unix,"
              "and how to compile it for macOS.")
        return

    df_proc = df.copy()
    old_columns = df_proc.columns
    df_proc.columns = df_proc.columns.get_level_values(level=0)
    df_proc.index = pd.to_datetime(df_proc.index, errors="coerce")

    trends = []
    seas_adjs = []
    for column in range(len(df_proc.columns)):
        series = df_proc.iloc[:, column].dropna()
        try:
            decomposition = tsa.x13_arima_analysis(
                series, outlier=outlier, trading=trading, forecast_periods=0,
                x12path=binary_path, prefer_x13=True
            )
            trend = decomposition.trend.reindex(df_proc.index)
            seas_adj = decomposition.seasadj.reindex(df_proc.index)

        except X13Error:
            if outlier is True:
                try:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}' with selected "
                          f"parameters. Trying with outlier=False...")
                    return decompose(df=df, outlier=False)

                except X13Error:
                    try:
                        print(f"X13 error found while processing "
                              f"'{df_proc.columns[column]}' with "
                              f"trading=True. Trying with trading=False...")
                        return decompose(df=df, outlier=False,
                                         trading=False)

                    except X13Error:
                        print(f"X13 error found while processing "
                              f"'{df_proc.columns[column]}'. "
                              f"Filling with nan.")
                        trend = pd.Series(np.nan, index=df_proc.index)
                        seas_adj = pd.Series(np.nan, index=df_proc.index)

            elif trading is True:
                try:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}' "
                          f"with trading=True. Trying with "
                          f"trading=False...")
                    return decompose(df=df, trading=False)

                except X13Error:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}'. Filling with nan.")
                    trend = pd.Series(np.nan, index=df_proc.index)
                    seas_adj = pd.Series(np.nan, index=df_proc.index)

            else:
                try:
                    return decompose(df=df)

                except X13Error:
                    print(f"X13 error found while processing "
                          f"'{df_proc.columns[column]}'. "
                          f"Filling with nan.")
                    trend = pd.Series(np.nan, index=df_proc.index)
                    seas_adj = pd.Series(np.nan, index=df_proc.index)

        trends.append(trend)
        seas_adjs.append(seas_adj)

    trends = pd.concat(trends, axis=1)
    seas_adjs = pd.concat(seas_adjs, axis=1)

    trends.columns = old_columns
    seas_adjs.columns = old_columns

    columns._setmeta(trends, seas_adj="Tendencia")

    columns._setmeta(seas_adjs, seas_adj="SA")

    return trends, seas_adjs
